#include<stdio.h>
#include<syslog.h>
#include"fdist.h"
#include<errno.h>
#include<signal.h>
#include"md5.h"

#define F_CONFILENAME	0x001
#define F_OUTFILENAME	0x002
#define F_NODETACH	0x004

static int debuglevel=0;
#define	DEBUG(n) if ( (n) <= debuglevel )

static struct MD5Context ctx ; /* MD5 info tu chexum the writen/read file */
static u_int options=0;
static u_int timeout=300;
static char gotdigest[16];
static char cntdigest[16];
static char filename[MAXFILENAMELEN];
static char szimname[MAXFILENAMELEN]="default";
static int filefd,s;
static unsigned short int sportnum=4444;
static struct sockaddr_in toaddr;
static struct hostent *he;
static struct timeval stt;
static konfentry *konf=NULL;
static char * cp;
static FILE * cfgf;
static char buff[CFGREADBUFLEN];
static int flag=1,wr=0;
static fd_set topoll;
static initstr inif;

static void usage (void);
static void client (void);
static void server (void);
static RETSIGTYPE cfinish (int parm); /* Szerver child processz kilpsi rutinja */
static RETSIGTYPE finish (int parm);  /* Szerver processz kilpsi rutinja */
static RETSIGTYPE konfreread (int parm); /* A szerver sighup-ra olvassa jra a konfigot */
static int ParseConfEntry (konfentry *k, char * buf); /* Szerver konfigurcis llomny feldolgozshoz val */
static void AddNewConfEntry (char * konfline); /* Szerver konfigurcis llomny feldolgozshoz val */
static konfentry * CreateConfEntry (char * buf); /* Konfig jraolvasshoz kell */
static void ShutConfEntry (konfentry *k); /* Kliens eldobsa */
static konfentry * AddNewClientConnect ();
void kliensreset (konfentry * k,int toclose);

int main( int argc, char * argv[]) {
  char ch;
  int i;
  konfentry *ce;
  while ((ch=getopt(argc,argv,"hnd:c:t:p:f:")) != EOF)
    switch(ch) {
      case 'p':
	sportnum=atoi(optarg);
	break;
      case 't':
	timeout=atoi(optarg);
	break;
      case 'd':
	debuglevel=atoi(optarg);
	break;
      case 'f':
	strncpy(filename,optarg,MAXFILENAMELEN);
	options |= F_OUTFILENAME;
	break;
      case 'c':
	strncpy(filename,optarg,MAXFILENAMELEN);
	options |= F_CONFILENAME;
	break;
      case 'n':
	options |= F_NODETACH;
	break;
      case 'h':
      default:
	usage();
	exit(2);
    }
  argc -=optind;
  argv +=optind;

  if (!argc) {
    fprintf(stderr,"%s: no hostname\n",PACKAGE);
    usage();
    exit(2);
  }

  if ( options & F_OUTFILENAME && options & F_CONFILENAME ) {
    fprintf(stderr,"%s: [-f outfilename|-c confilename]\nParameter conflict.\n",PACKAGE);
    usage();
    exit(2);
  }

  if (!(options & F_CONFILENAME)) { /* Kliens kd init */
    if(strchr(*argv,'@')) { /* Definilt a szimbolnv */
      if((cp=strtok(*argv,"@"))){
	strncpy(szimname,cp,MAXFILENAMELEN);
	cp=strtok(NULL,"\n");
	if(!cp) {
	  usage();
	  exit(2);
	}
      } else {
	fprintf(stderr,"%s: No szimbolname\n",PACKAGE);
	usage();
      }
    } else { /* nem definilt a szimbolnv, marad a default */
      cp=*argv;
    }
    memset(&toaddr,0,sizeof(struct sockaddr));
    toaddr.sin_family=AF_INET;
    if(inet_aton(cp,&toaddr.sin_addr)) {
    } else {
      he=gethostbyname(cp);
      if(!he) {
	fprintf(stderr,"%s: unknown host %s\n",PACKAGE,cp);
	exit(2);
      }
      toaddr.sin_family=he->h_addrtype;
      if(he->h_length>(int)sizeof(toaddr.sin_addr)){
	he->h_length=sizeof(toaddr.sin_addr);
      }
      memcpy(&toaddr.sin_addr,he->h_addr,he->h_length);
    }
    toaddr.sin_port=htons(sportnum);
    if ((s=socket(AF_INET,SOCK_STREAM,0))<0) {
      perror(PACKAGE ": socket");
      exit(2);
    }
    if (connect(s,(struct sockaddr*)&toaddr,sizeof(struct sockaddr))) {
      perror(PACKAGE ": connect");
      exit(2);
    }
    if (options & F_OUTFILENAME) {
      if((filefd=open(filename,O_WRONLY|O_CREAT))==-1) {
	perror(PACKAGE ": open output file");
	exit(3);
      }
    } else { /* Ha nincs megadva mi legyen a kimeneti fjl, akkor stdout */
      filefd=1;
    }
    MD5Init(&ctx);
    client();
    MD5Final(cntdigest,&ctx);
    read(s,&gotdigest,16);
    if(memcmp(cntdigest,gotdigest,16)) {
      fprintf(stderr,"%s: counted and got MD5sums mismatch!\n",PACKAGE);
      exit(2);
    }
  } else { /* Szerver kd init */
    if (!(cfgf=fopen(filename,"r"))) {
      perror(PACKAGE ": open config file");
      exit(3);
    }
    while(!feof(cfgf)) {
      if(fgets(buff,CFGREADBUFLEN,cfgf))
	AddNewConfEntry(buff);
    }
    fclose(cfgf);
    signal(SIGINT,finish);
    signal(SIGTERM,finish);
    signal(SIGHUP,konfreread);
    openlog(PACKAGE,LOG_PID,LOG_DAEMON|LOG_NOTICE|LOG_FACMASK);
    syslog(LOG_DAEMON|LOG_NOTICE,"%s server started",PACKAGE);
    i=0; ce=konf;
    while (ce) {
      i++;
      ce=(konfentry *) ce->next;
    }
    syslog(LOG_DAEMON|LOG_NOTICE,"%s: %d config lines parsed",PACKAGE,i);
    if ((s=socket(AF_INET,SOCK_STREAM,0))<0) {
      perror(PACKAGE ": socket");
      syslog(LOG_DAEMON|LOG_NOTICE,"%s couldnt create socket",PACKAGE);
      exit(2);
    }
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(flag))==-1) {
      perror(PACKAGE ": setsockopt");
      syslog(LOG_DAEMON|LOG_NOTICE,"%s couldnt set SO_REUSEADDR",PACKAGE);
      exit(2);
    }
    toaddr.sin_family=AF_INET;
    toaddr.sin_addr.s_addr=INADDR_ANY;
    toaddr.sin_port=htons(sportnum);
    if (bind(s,(struct sockaddr*)&toaddr,sizeof(toaddr))==-1) {
      perror(PACKAGE "-listener: bind");
      syslog(LOG_DAEMON|LOG_NOTICE,"%s: couldnt bind",PACKAGE);
      exit(2);
    }
    /* A ktelez argumentum a szrcm */
    if(! inet_aton(*argv,&toaddr.sin_addr)) {
      fprintf(stderr,"ktelez brdkaszt cm kell\n");
      usage();
      exit(2);
    }
    if (! (options & F_NODETACH)) {
      daemon(0,0);
      syslog(LOG_DAEMON|LOG_NOTICE,"%s: daemonized self",PACKAGE);
    }
    if(listen(s,0)) {
      perror(PACKAGE ": listen");
      syslog(LOG_DAEMON|LOG_NOTICE,"%s couldnt start listen",PACKAGE);
      exit(2);
    }
    while (1) {
      FD_ZERO(&topoll);
      FD_SET(s,&topoll);
      stt.tv_sec=1;
      stt.tv_usec=0;
      if (select(s+1,&topoll,NULL,NULL,&stt)==-1) {
	perror(PACKAGE ": select");
	if (wr) wr=0; else exit(4);
      }
      if (FD_ISSET(s,&topoll)) {
	if ((ce=AddNewClientConnect())) {
	  if ((i=fork())<0) {
	    perror(PACKAGE ": fork");
	  } else if (i==0) {
	    signal(SIGINT,cfinish);
	    signal(SIGTERM,cfinish);
	    signal(SIGHUP,cfinish);
	    konf=ce;
	    server();
	    exit(0);
	  } else {
	    kliensreset(ce,1);
	    ce->szim.childpid=i;
            syslog(LOG_DAEMON|LOG_NOTICE,"%s child[%d] started(%s:%d)",PACKAGE,i,ce->szimbol,ce->toreach);
	  }
	}
      } else sleep(1);
      /* Gyerektakarts */
      ce=konf;
      while (ce) {
	if (ce->szim.childpid==-1) {ce=ce->next; continue;}
	if (waitpid(ce->szim.childpid,NULL,WNOHANG)) {
	  /* kliensreset(ce); Ezt elmletileg mr a fork() utn kellene */
          syslog(LOG_DAEMON|LOG_NOTICE,"%s child[%d] stopped",PACKAGE,ce->szim.childpid);
	  ce->szim.childpid=-1;
	}
	ce=ce->next;
      }
    }
  }
  return 0;
}

void usage (void) {
  fprintf(stderr,"Usage:\nfdist [-p serverport] [-f outfilename|-c configfilename]"
		    "\n\t[-t timeout] [-d debuglevel] [-n] [szimbol@]targethost\n"
		    "%s: v %s\n",PACKAGE,VERSION);
}

int ParseConfEntry (konfentry * k, char * buf) {
  char * cp;
  /* szimblumnv beolvassa */
  cp=strtok(buf,":");
  if(cp) strncpy(k->szimbol,cp,MAXFILENAMELEN);
  else return 1;
  /* fjlnv beolvassa */
  cp=strtok(NULL,":");
  if(cp) strncpy(k->filenev,cp,MAXFILENAMELEN);
  else return 1;
  /* kliensszm beolvassa */
  cp=strtok(NULL,":");
  if(cp) k->toreach=atoi(cp);
  else return 1;
  /* ablakszm beolvassa */
  cp=strtok(NULL,":");
  if(cp) k->szim.ablaknum=atoi(cp);
  else return 1;
  /* ablakmret beolvassa */
  cp=strtok(NULL,":");
  if(cp) k->szim.blocksize=atoi(cp);
  else return 1;
  if (k->szim.blocksize>=SSIZE_MAX) return 2;
  /* adport beolvassa */
  cp=strtok(NULL,":");
  if(cp) k->szim.sendport=atoi(cp);
  else return 1;
  /* fogadport beolvassa */
  cp=strtok(NULL,":");
  if(cp) k->szim.recvport=atoi(cp);
  else return 1;
  while (strtok(NULL,":")) ;
  return 0;
}

void AddNewConfEntry (char * buf) {
  int retc;
  konfentry * k;
  if(buf[0]=='#') { return; }
  k=malloc(sizeof(konfentry));
  k->next=konf;
  if((retc=ParseConfEntry(k,buf))) {
    DEBUG(2) switch(retc) {
      case 2:
	syslog(LOG_DAEMON|LOG_NOTICE,"%s: hibs konfigsor blocksize %d >= %d \n",PACKAGE,k->szim.blocksize,SSIZE_MAX);
	break;
      case 1:
      default:
	syslog(LOG_DAEMON|LOG_NOTICE,"%s: hibs konfigsor: %s\n",PACKAGE,buf);
	break;
    }
    free(k);
    return;
  }
  k->szim.childpid=-1;
  k->szim.numclients=0;
  k->szim.clients=calloc(k->toreach,sizeof(clientinfo));
  if(!(k->szim.clients)) {
    free(k); DEBUG(2) syslog(LOG_DAEMON|LOG_NOTICE,"%s: nem tudtam kliensinfot allokalni\n",PACKAGE); return;
  }
  kliensreset(k,0);
  konf=k;
}

konfentry * CreateConfEntry (char * buf) {
  int retc;
  konfentry * k;
  if(buf[0]=='#') { return NULL; }
  k=malloc(sizeof(konfentry));
  k->next=konf;
  DEBUG(5) syslog(LOG_DAEMON|LOG_NOTICE,"%s: konfigsor: %s\n",PACKAGE,buf);
  if((retc=ParseConfEntry(k,buf))) {
    DEBUG(2) switch(retc) {
      case 2:
	syslog(LOG_DAEMON|LOG_NOTICE,"%s: hibs konfigsor blocksize %d >= %d \n",PACKAGE,k->szim.blocksize,SSIZE_MAX);
	break;
      case 1:
      default:
	syslog(LOG_DAEMON|LOG_NOTICE,"%s: hibs konfigsor: %s\n",PACKAGE,buf);
	break;
    }
    free(k);
    return NULL;
  }
  k->szim.childpid=-1;
  k->szim.numclients=0;
  k->wasrenewed=1;
  k->szim.clients=calloc(k->toreach,sizeof(clientinfo));
  if(!(k->szim.clients)) {
    free(k); DEBUG(2) fprintf(stderr,"%s: nem tudtam kliensinfot allokalni\n",PACKAGE); return NULL;
  }
  kliensreset(k,0);
  return k;
}

void ShutConfEntry (konfentry * k) {
  if(k->szim.childpid!=-1)
    syslog(LOG_DAEMON|LOG_NOTICE,"%s child[%d] detached",PACKAGE,k->szim.childpid);
  kliensreset(k,1);
  free(k->szim.clients);
  free(k);
}

static konfentry * AddNewClientConnect () {
  int newfd;
  int x;
  konfentry *ce;
  newfd=accept(s,NULL,NULL);
  x=read(newfd,szimname,MAXFILENAMELEN);
  if (!x) {
    usleep(100);
    x=read(newfd,szimname,MAXFILENAMELEN);
  }
  if (!x) {
    syslog(LOG_DAEMON|LOG_NOTICE,"client timed out");
    close(newfd);
    return NULL;
  }
  ce=konf;
  while(ce) {
    if (!strncmp(szimname,ce->szimbol,MAXFILENAMELEN)) break;
    ce=ce->next;
  }
  if (ce) {
    if (ce->szim.childpid!=-1) {
      close(newfd);
      syslog(LOG_DAEMON|LOG_NOTICE,"refused client, child process already working");
      return NULL;
    }
    ce->szim.clients[ce->szim.numclients]=newfd;
    (ce->szim.numclients)++;
    if ((ce->szim.numclients)==(ce->toreach)) return ce;
  }
  if (!ce) {
    close(newfd);
    syslog(LOG_DAEMON|LOG_NOTICE,"tried to get nonexistant %s szimbol",szimname);
  }
  return NULL;
}

void kliensreset (konfentry * k,int toclose) {
  int i;
  for(i=0;i< k->toreach;i++) {
    if (toclose && ((k->szim.clients)[i] !=-1 )) close((k->szim.clients)[i]);
    (k->szim.clients)[i]=-1;
  }
  k->szim.numclients=0;
}

void client() {
  int recvfd;
  int *ind;
  int i,ungetno,n=0;
  long ofszet=0;
  long sent_no=-1;
  long getwinnum;
  window * ablak;
  DEBUG(2) fprintf(stderr,"kld %s\n",szimname);
  write(s,szimname,strlen(szimname)+1);
  if(read(s,&inif,sizeof(inif))!=sizeof(inif)) {
    fprintf(stderr,"incompatible server type or other error\n"); /* Oder error lehet az is, ha nincs olyan szimbol a szerveren, mint amilyet krtnk! */
    exit(5);
  }
  DEBUG(5) fprintf(stderr,"inicializl adatok jttek ablakszm=%d,sendport=%d,recvport=%d\n",ntohs(inif.ablaknum),ntohs(inif.sendport),ntohs(inif.recvport));
  /* ind tbla fellltsa */
  if(!(ind=calloc(ntohs(inif.ablaknum)+1,sizeof(int)))) {
    /* egyel tbbet hagyunk, ha neadjisten pluszba
     * bekapunk mg vmi csomagot,
     * avval is el tudjunk bnni. */
    perror(PACKAGE ": ind calloc");
    exit(2);
  }
  for (i=0;i<=ntohs(inif.ablaknum);i++)ind[i]=-1;
  /* ablakok lefoglalsa */
  if(!(ablak=calloc(ntohs(inif.ablaknum)+1,sizeof(window)))) {
    perror(PACKAGE ": win calloc");
    exit(2);
  }
  /* recvfd setupolsa! */
  if ((recvfd=socket(AF_INET,SOCK_DGRAM,0))<0) {
    perror(PACKAGE ": socket");
    exit(2);
  }
  toaddr.sin_family=AF_INET;
  toaddr.sin_addr.s_addr=INADDR_ANY;
  toaddr.sin_port=inif.recvport;
  if (bind(recvfd,(struct sockaddr*)&toaddr,sizeof(toaddr))==-1) {
    perror(PACKAGE "-client: bind");
    DEBUG(1) fprintf(stderr,"tried to bind to port %d",ntohs(inif.recvport));
    exit(2);
  }
  if (setsockopt(recvfd,SOL_SOCKET,SO_BROADCAST,(char *)&flag,sizeof(flag))) {
    perror(PACKAGE ": broadcast");
  }
  /* client fciklus */
  while(sent_no) {
    FD_ZERO(&topoll);
    FD_SET(s,&topoll);
    FD_SET(recvfd,&topoll);
    stt.tv_sec=timeout;
    stt.tv_usec=0;
    switch(select(MAX(s,recvfd)+1,&topoll,NULL,NULL,&stt)) {
      case 0:
      case -1:
	fprintf(stderr,"error occured during select\n");
	exit(5);
	break;
      default:
	if(FD_ISSET(recvfd,&topoll)) {
	  i=read(recvfd,&(ablak[n].winnum),sizeof(window));
	  DEBUG(7) fprintf(stderr,"adat jtt winnum=%ld n=%d\n",ablak[n].winnum,n);
	  if (i!=(sizeof(long)+sizeof(short)+ablak[n].datalen))
	    fprintf(stderr,"winnum size and read data mismatch\n");
	  if (ablak[n].winnum-ofszet>=0 &&
	      ablak[n].winnum-ofszet<inif.ablaknum)
	    if (ind[ablak[n].winnum-ofszet]==-1) {
	      ind[ablak[n].winnum-ofszet]=n;
	      n++;
	    }
	  break;
	} /* brdkasztba kapott adat lekezelse */
	if(FD_ISSET(s,&topoll)) {
	  read(s,&sent_no,sizeof(sent_no));
	  sent_no=ntohl(sent_no);
	  DEBUG(6) fprintf(stderr,"kontrol-adat jtt %ld\n",sent_no);
	  getwinnum=ofszet+inif.ablaknum;
	  if(sent_no) {
	    for(ungetno=0;ungetno<inif.ablaknum;ungetno++)
	      if (ind[ungetno]==-1) { getwinnum=ofszet+ungetno; break; }
	  }
	  if ( getwinnum==sent_no || (!sent_no)) {
	    for(i=0;i<sent_no-ofszet;i++) {
	      if(ablak[ind[i]].datalen>0) {
		MD5Update(&ctx,(md5byte *)&(ablak[ind[i]].data),ablak[ind[i]].datalen);
		write(filefd,&(ablak[ind[i]].data),ablak[ind[i]].datalen);
	      }
	      ind[i]=-1;
	      n=0;
	    }
	    ofszet=getwinnum;
	  }
	  getwinnum=htonl(getwinnum);
	  write(s,&getwinnum,sizeof(getwinnum));
	  DEBUG(6) fprintf(stderr,"kontrol-adat klds %d\n",ntohl(getwinnum));
	} /* kontrol csatornn jtt adat lekezelse */
    } /* switch */
  }
}

void server() {
  window * ablak;
  int i,sendfd,htr;
  long sent_no=0,ofszet,n,eof=0;
  if((filefd=open(konf->filenev,O_RDONLY))==-1) {
    syslog(LOG_DAEMON|LOG_NOTICE,"%s: open input file error",PACKAGE);
    for (i=0;i<konf->szim.numclients;i++)
	close(konf->szim.clients[i]);
    exit(3);
  }
  inif.recvport=htons(konf->szim.recvport);
  inif.sendport=htons(konf->szim.sendport);
  inif.ablaknum=htons(konf->szim.ablaknum);
  for(i=0;i<konf->szim.numclients;i++) 
    write(konf->szim.clients[i],&inif,sizeof(inif));
  MD5Init(&ctx);
  /* ablakok foglalsa */
  inif.ablaknum=ntohs(inif.ablaknum);
  if(!(ablak=calloc(inif.ablaknum,sizeof(window)))) {
    syslog(LOG_DAEMON|LOG_NOTICE,"%s: win calloc",PACKAGE);
    exit(2);
  }
  /* sendfd setupolsa */
  if ((sendfd=socket(AF_INET,SOCK_DGRAM,0))<0) {
    perror(PACKAGE ": socket");
    exit(2);
  }
  if (setsockopt(sendfd,SOL_SOCKET,SO_BROADCAST,(char *)&flag,sizeof(flag))) {
    perror(PACKAGE ": broadcast");
  }
  /* toaddr.sin_addr -t mg a szl kitlttte, a brdkaszt cmmel */
  toaddr.sin_family=AF_INET;
  toaddr.sin_port=inif.sendport;
  if (bind(sendfd,(struct sockaddr*)&toaddr,sizeof(toaddr))==-1) {
    perror(PACKAGE "-sender: bind");
    DEBUG(1) fprintf(stderr,"%s tried to bind to port %d",PACKAGE,ntohs(inif.sendport));
    exit(2);
  }
  toaddr.sin_port=inif.recvport;
  while(!eof) {
    ofszet=ntohl(sent_no);
    for(i=0;i<inif.ablaknum;i++) {
      /* elkszt ablak i */
      DEBUG(7) fprintf(stderr,PACKAGE "-sender: kld %d wind\n",i);
      ablak[i].winnum=ofszet+i;
      ablak[i].datalen=read(filefd,ablak[i].data,konf->szim.blocksize);
      if (ablak[i].datalen<konf->szim.blocksize) eof=ablak[i].winnum+1;
      if (ablak[i].datalen==0) eof=ablak[i].winnum;
      /* sztszr ablak i */
      sendto(sendfd,&(ablak[i].winnum),((char*)&(ablak[i].data[ablak[i].datalen]))-((char*)&(ablak[i].winnum)),
	  MSG_DONTROUTE,(struct sockaddr*)&toaddr,sizeof(toaddr));
      /* MD5frisst */
      MD5Update(&ctx,ablak[i].data,ablak[i].datalen);
      if(eof) break;
    }
    if (eof) sent_no=eof;
    else sent_no=ofszet+inif.ablaknum;
    sent_no=htonl(sent_no);
    do {
      htr=0;
      for (i=0;i<konf->szim.numclients;i++)
	write(konf->szim.clients[i],&sent_no,sizeof(sent_no));
      DEBUG(7) fprintf(stderr,"kliensek kirtestse %d\n",ntohl(sent_no));
      for (i=0;i<inif.ablaknum;i++)
	ablak[i].was_resent=0;
      i=0; while (i<konf->szim.numclients) { /* for i=0;i<kliensnum;i++ */
	if ((read(konf->szim.clients[i],&n,sizeof(n)))!=sizeof(n)) { /* kliens ledgltt, vagy megtvedt */
	  syslog(LOG_DAEMON|LOG_NOTICE,"eltvedt klienst kigyilkolok");
	  for(n=i;n<konf->szim.numclients-1;n++)
	    memcpy(&(konf->szim.clients[i]),&(konf->szim.clients[i+1]),sizeof(konf->szim.clients[i]));
	  konf->szim.numclients--;
	  if(konf->szim.numclients==0) {
	    syslog(LOG_DAEMON|LOG_NOTICE,"a kliensek elfogytak :-)");
	    DEBUG(5) fprintf(stderr,"a kliensek elfogytak :-)");
	    exit(2);
	  }
	} else {
	  n=ntohl(n);
	  if (ntohl(sent_no)>n) {
	    /* ide jn az jraklds */
	    DEBUG(5) fprintf(stderr,"jrakldm %ld windt(%d)\n",n,ntohl(sent_no));
	    n=n-ofszet;
	    DEBUG(5) fprintf(stderr,"tszmtott rtke: %ld\n",n);
	    //if (! ablak[n].was_resent) {
	      sendto(sendfd,&(ablak[n].winnum),((char*)&(ablak[n].data[ablak[n].datalen]))-((char*)&(ablak[n].winnum)),
		  MSG_DONTROUTE,(struct sockaddr*)&toaddr,sizeof(toaddr));
	      ablak[n].was_resent=1;
	    //}
	    htr=1;
	  }
	}
	i++; /* for->while konverzi miatt */
      }
    } while (htr);/* do-while */
  } /* fjl vgig klds */
  DEBUG(5) fprintf(stderr,"Fjl vgig elkldve\n");
  n=htonl(0); /* 0 a fjl vge jel, ezt kldjk, mert a vgre rtnk */
  for (i=0;i<konf->szim.numclients;i++)
    write(konf->szim.clients[i],&n,sizeof(n));
  MD5Final(cntdigest,&ctx);
  for (i=0;i<konf->szim.numclients;i++)
    write(konf->szim.clients[i],cntdigest,16);
  DEBUG(5) fprintf(stderr,"server() end\n");
}

static RETSIGTYPE cfinish (int parm) { /* A szerver ltal indtott gyerekek gykell kilpjenek */
  syslog(LOG_DAEMON|LOG_NOTICE,"%s server child exited",PACKAGE);
  closelog();
  exit(0);
}

static RETSIGTYPE konfreread(int parm) { /* A szerver sighupra olvassa jra a konfigfjlt */
  konfentry *k,*l;
  int db;
  if (!(cfgf=fopen(filename,"r"))) {
    syslog(LOG_DAEMON|LOG_NOTICE,"%s: open config file",PACKAGE);
    exit(3);
  }
  k=konf;
  while(k) {
    k->wasrenewed=0;
    k=k->next;
  }
  while(!feof(cfgf)) {
    if(fgets(buff,CFGREADBUFLEN,cfgf)) {
      k=CreateConfEntry(buff);
      l=konf;
      if (k) {
	while(l) {
	  if(!strncmp(l->szimbol,k->szimbol,MAXFILENAMELEN))
	    break;
	  l=l->next;
	}
      } else l=NULL;
      if(l) {
	l->wasrenewed=1;
	if(l->szim.childpid == -1) { /* Ha nincs gyerek forkolva */
	  if (l->szim.numclients>0)
	    syslog(LOG_DAEMON|LOG_NOTICE,"%s: bunkn lelvk kapcsold gyerekeket (%s) konfreread miatt",PACKAGE,k->szimbol);
	  kliensreset(l,1);
	} else { /* Ha pp fork()-olva van egy child */
	  l->szim.numclients=0;
	}
	l->szim.recvport=k->szim.recvport;
	l->szim.ablaknum=k->szim.ablaknum;
	l->szim.blocksize=k->szim.blocksize;
	l->toreach=k->toreach;
	strncpy(l->filenev,k->filenev,MAXFILENAMELEN);
	ShutConfEntry(k);
      } else {
	if (k) konf=k;
      }
    }

  }
  fclose(cfgf);
  syslog(LOG_DAEMON|LOG_NOTICE,"%s confreread complete",PACKAGE);
  while(konf && (!konf->wasrenewed)) { l=konf->next; ShutConfEntry(konf); konf=l; } /* Pusztuljon a frgese */
  l=konf;
  k=konf->next;
  while(k) {
    if(k->wasrenewed) {
      l=k;
    } else {
      l->next=k->next;
      ShutConfEntry(konf);
    }
    k=k->next;
  }
  db=0; k=konf; while (k) { db++; k=k->next; }
  syslog(LOG_DAEMON|LOG_NOTICE,"%s: in config %d lines was reread",PACKAGE,db);
  wr=1;
  return;
}

static RETSIGTYPE finish (int parm) { /* A szerver ki tudjon lpni */
  konfentry * ce;
  ce=konf;
  while(ce) {
    konf=konf->next;
    free(ce->szim.clients);
    free(ce);
    ce=konf;
  }
  syslog(LOG_DAEMON|LOG_NOTICE,"%s server stopped",PACKAGE);
  closelog();
  exit(0);
}

